(***********************************************************************)
(*                                                                     *)
(*                               TAToo                                 *)
(*                                                                     *)
(*                     Kim Nguyen, LRI UMR8623                         *)
(*                   Université Paris-Sud & CNRS                       *)
(*                                                                     *)
(*  Copyright 2010-2013 Université Paris-Sud and Centre National de la *)
(*  Recherche Scientifique. All rights reserved.  This file is         *)
(*  distributed under the terms of the GNU Lesser General Public       *)
(*  License, with the special exception on linking described in file   *)
(*  ../LICENSE.                                                        *)
(*                                                                     *)
(***********************************************************************)

INCLUDE "utils.ml"
INCLUDE "debug.ml"

open Format
open Misc
open Bigarray

type stats = { mutable pass : int;
               tree_size : int;
               mutable fetch_trans_cache_access : int;
               mutable fetch_trans_cache_miss : int;
               mutable eval_trans_cache_access : int;
               mutable eval_trans_cache_miss : int;
               mutable nodes_per_run : int list;
             }

module NodeSummary =
struct
  (* Pack into an integer the result of the is_* and has_ predicates
     for a given node *)
  type t = int
  let dummy = -1
  (*
    ...44443210
    ...4444 -> kind
    3 -> has_right
    2 -> has_left
    1 -> is_right
    0 -> is_left
  *)
  let is_left (s : t) : bool =
    s land 1 != 0

  let is_right (s : t) : bool =
    s land 0b10 != 0

  let has_left (s : t) : bool =
    s land 0b100 != 0

  let has_right (s : t) : bool =
    s land 0b1000 != 0

  let kind (s : t) : Tree.NodeKind.t =
    Obj.magic (s lsr 4)

  let make is_left is_right has_left has_right kind =
    (int_of_bool is_left) lor
      ((int_of_bool is_right) lsl 1) lor
      ((int_of_bool has_left) lsl 2) lor
      ((int_of_bool has_right) lsl 3) lor
      ((Obj.magic kind) lsl 4)
end

let dummy_set = StateSet.singleton State.dummy_state



  IFDEF HTMLTRACE
  THEN
type sat_array = StateSet.t array list
  DEFINE IFHTML(a,b) = (a)
  ELSE
type sat_array = StateSet.t array
  DEFINE IFHTML(a,b) = (b)
  END

let unsafe_get a i =
  if i < 0 then StateSet.empty else
    Array.unsafe_get (IFHTML(List.hd a, a)) i

let unsafe_set a i v old_v =
  if v != old_v then
    Array.unsafe_set (IFHTML(List.hd a, a)) i v

type 'a run = {
  tree : 'a ;
     (* The argument of the run *)
  auto : Ata.t;
     (* The automaton to be run *)
  mutable sat: sat_array;
     (* A mapping from node preorders to states satisfied at that node *)
  mutable pass : int;
     (* Number of run we have performed *)
  mutable fetch_trans_cache : Ata.Formula.t Cache.N2.t;
     (* A cache from states * label to list of transitions *)
  mutable td_cache : StateSet.t Cache.N6.t;
  mutable bu_cache : StateSet.t Cache.N6.t;
     (* Two 6-way caches used during the top-down and bottom-up phase
        label * self-set * fc-set * ns-set * parent-set * node-shape -> self-set
     *)
  node_summaries: (int, int16_unsigned_elt, c_layout) Array1.t;
  stats : stats;
}

let dummy_form = Ata.Formula.stay State.dummy_state

let get_form run tag (q : State.t) =
  let auto = run.auto in
  let fetch_trans_cache = run.fetch_trans_cache in
  let stats = run.stats in
  let phi =
    stats.fetch_trans_cache_access <- stats.fetch_trans_cache_access + 1;
    Cache.N2.find fetch_trans_cache (tag.QName.id :> int) (q :> int)
  in
  if phi == dummy_form then
    let phi = Ata.get_form auto tag q in
    let () =
      stats.fetch_trans_cache_miss <- stats.fetch_trans_cache_miss + 1;
      Cache.N2.add
        fetch_trans_cache
        (tag.QName.id :> int)
        (q :> int) phi
    in phi
  else
    phi



let eval_form phi fcs nss ps ss summary =
  let open Ata in
  let rec loop phi =
    begin match Formula.expr phi with
    | Boolean.False -> false
    | Boolean.True -> true
    | Boolean.Atom (a, b) ->
      begin
        let open NodeSummary in
        match a.Atom.node with
        | Move (m, q) ->
          b && StateSet.mem q (
            match m with
              `First_child -> fcs
            | `Next_sibling -> nss
            | `Parent | `Previous_sibling -> ps
            | `Stay -> ss
          )
        | Is_first_child -> b == is_left summary
        | Is_next_sibling -> b == is_right summary
        | Is k -> b == (k == kind summary)
        | Has_first_child -> b == has_left summary
        | Has_next_sibling -> b == has_right summary
      end
    | Boolean.And(phi1, phi2) -> loop phi1 && loop phi2
    | Boolean.Or (phi1, phi2) -> loop phi1 || loop phi2
    end
  in
  loop phi


let eval_trans_aux run tag summary fcs nss ps sat todo  =
  StateSet.fold (fun q (a_sat) ->
    let phi =
      get_form run tag q
    in
    if eval_form phi fcs nss ps a_sat summary then
      StateSet.add q a_sat
    else a_sat
  ) todo sat


let rec eval_trans_fix run tag summary fcs nss ps sat todo  =
  let new_sat =
    eval_trans_aux run tag summary fcs nss ps sat todo
  in
  if new_sat == sat then sat else
    eval_trans_fix run tag summary fcs nss ps new_sat todo


let eval_trans run trans_cache tag summary fcs nss ps ss todo =
  let stats = run.stats in
  let fcsid = (fcs.StateSet.id :> int) in
  let nssid = (nss.StateSet.id :> int) in
  let psid = (ps.StateSet.id :> int) in
  let ssid = (ss.StateSet.id :> int) in
  let tagid = (tag.QName.id :> int) in

  let res = Cache.N6.find trans_cache tagid summary ssid fcsid nssid psid in
  stats.eval_trans_cache_access <- 1 + stats.eval_trans_cache_access;
  if res != dummy_set then
    res
  else let new_sat =
             eval_trans_fix run tag summary fcs nss ps ss todo
    in
    stats.eval_trans_cache_miss <- 1 + stats.eval_trans_cache_miss;
    Cache.N6.add trans_cache tagid summary ssid fcsid nssid psid new_sat;
    new_sat


module Make (T : Tree.S) (L : Deque.S with type elem = T.node) =
struct

  let make auto tree =
    let len = T.size tree in
    let ba = Array1.create int16_unsigned c_layout len in
    Array1.fill ba 0;
    {
      tree = tree;
      auto = auto;
      sat = (let a = Array.make len StateSet.empty in
             IFHTML([a], a));
      pass = 0;
      fetch_trans_cache = Cache.N2.create dummy_form;
      td_cache = Cache.N6.create dummy_set;
      bu_cache = Cache.N6.create dummy_set;
      node_summaries = ba;
      stats = {
        pass = 0;
        tree_size = len;
        fetch_trans_cache_access = 0;
        fetch_trans_cache_miss = 0;
        eval_trans_cache_access = 0;
        eval_trans_cache_miss = 0;
        nodes_per_run = [];
      }
    }


  let top_down run update_res =
    let num_visited = ref 0 in
    let i = run.pass in
    let tree = run.tree in
    let auto = run.auto in
    let states_by_rank = Ata.get_states_by_rank auto in
    let td_todo = states_by_rank.(i) in
    let bu_todo =
      if i == Array.length states_by_rank - 1 then StateSet.empty
      else
        states_by_rank.(i+1)
    in
    let last_run = i >= Array.length states_by_rank - 2 in
    let rec loop_td_and_bu node parent parent_sat =
      if node == T.nil then StateSet.empty
      else begin
        incr num_visited;
        let tag = T.tag tree node in
        let node_id = T.preorder tree node in
        let fc = T.first_child tree node in
        let ns = T.next_sibling tree node in
         (* We enter the node from its parent *)
        let summary =
          let s = Array1.unsafe_get run.node_summaries node_id in
          if  s != 0 then s else
            let s =
              NodeSummary.make
                (node == (T.first_child tree parent)) (*is_left *)
                (node == (T.next_sibling tree parent)) (*is_right *)
                (fc != T.nil) (* has_left *)
                (ns != T.nil) (* has_right *)
                (T.kind tree node) (* kind *)
            in
            run.node_summaries.{node_id} <- s; s
        in
        let status0 = unsafe_get run.sat node_id in
        (* get the node_statuses for the first child, next sibling and parent *)
        (* evaluate the transitions with all this statuses *)
        let status1 =
          eval_trans run
            run.td_cache tag
            summary
            (unsafe_get run.sat (T.preorder tree fc))
            (unsafe_get run.sat (T.preorder tree ns))
            parent_sat
            status0 td_todo
        in
          (* update the cache if the status of the node changed
             unsafe_set run.sat node_id status1 status0;*)
        if bu_todo == StateSet.empty then begin
          unsafe_set run.sat node_id status1 status0; (* write the td_states *)
          update_res false status1 node;
          let _ = loop_td_and_bu fc node status1 in
          loop_td_and_bu ns node status1 (* tail call *)
        end else
          let fcs1, nss1 =
            if last_run then
              let nss1 = loop_td_and_bu ns node status1 in
              let fcs1 = loop_td_and_bu fc node status1 in
              fcs1, nss1
            else
              let fcs1 = loop_td_and_bu fc node status1 in
              let nss1 = loop_td_and_bu ns node status1 in
              fcs1, nss1
          in
          let status2 =
            eval_trans run run.bu_cache tag
              summary fcs1
              nss1
              parent_sat
              status1 bu_todo
          in
          unsafe_set run.sat node_id status2 status0;
          if last_run && status2 != StateSet.empty then update_res true status2 node;
          status2
      end
    in
    let _ = loop_td_and_bu (T.root tree) T.nil dummy_set in
    run.pass <- run.pass + 2;
    run.stats.pass <- run.stats.pass + 1;
    run.stats.nodes_per_run <- !num_visited :: run.stats.nodes_per_run



  let mk_update_result auto =
    let sel_states = Ata.get_selecting_states auto in
    let res = L.create () in
    (fun prepend sat node ->
      if StateSet.intersect sel_states sat then begin
        if prepend then L.push_front node res else
          L.push_back node res
      end),
    (fun () -> res)


  let mk_update_full_result auto =
    let dummy = L.create () in
    let res_mapper = Cache.N1.create dummy in
    let () =
      StateSet.iter
        (fun q -> Cache.N1.add res_mapper (q :> int) (L.create()))
        (Ata.get_selecting_states auto)
    in
    (fun prepend sat node ->
      StateSet.iter
        (fun q ->
          let res = Cache.N1.find res_mapper (q :> int) in
          if res != dummy then begin
            if prepend then L.push_front node res
            else L.push_back node res
          end
        ) sat),
    (fun () ->
      StateSet.fold_right
        (fun q acc -> (q, Cache.N1.find res_mapper (q :> int))::acc)
        (Ata.get_selecting_states auto) [])

  let prepare_run run list =
    let tree = run.tree in
    let auto = run.auto in
    let sat = IFHTML((List.hd run.sat), run.sat) in
    let sat0 = Ata.get_starting_states auto in
    L.iter (fun node ->
      let node_id = T.preorder tree node in
      sat.(node_id) <- sat0) list


  let compute_run auto tree nodes update_res =
    let run = make auto tree in
    prepare_run run nodes;
    let rank = Ata.get_max_rank auto in
    while run.pass <= rank do
      top_down run update_res;
      IFHTML((run.sat <- (Array.copy (List.hd run.sat)) :: run.sat), ());
      run.td_cache <- Cache.N6.create dummy_set;
      run.bu_cache <- Cache.N6.create dummy_set;
    done;
    IFHTML((run.sat <- List.tl run.sat), ());
    IFHTML(Html_trace.gen_trace auto run.sat (module T : Tree.S with type t = T.t) tree ,());
    run


  let last_stats = ref None

  let full_eval auto tree nodes =
    let update_full,get_full = mk_update_full_result auto in
    let run = compute_run auto tree nodes update_full in
    last_stats := Some run.stats;
    get_full ()

  let eval auto tree nodes =
    let update_res,get_res = mk_update_result auto in
    let run = compute_run auto tree nodes update_res in
    last_stats := Some run.stats;
    get_res ()

  let stats () = match !last_stats with
      Some s -> s.nodes_per_run <- List.rev s.nodes_per_run;s
    | None -> failwith "Missing stats"

end
